/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2020 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package dk.nsp.mingraviditet.ui.main.overview

import androidx.lifecycle.*
import com.commonsense.android.kotlin.base.extensions.coroutines.launchIO
import dagger.hilt.android.lifecycle.HiltViewModel
import dk.nsp.domain.knowledge.Article
import dk.nsp.domain.repository.ArticlesRepository
import dk.nsp.domain.repository.DataCardRepository
import dk.nsp.mingraviditet.R
import dk.nsp.mingraviditet.domain.AlertDialogButtonModel
import dk.nsp.mingraviditet.domain.AlertDialogViewModel
import dk.nsp.mingraviditet.extensions.fullName
import dk.nsp.mingraviditet.extensions.getOverviewPregnancyState
import dk.nsp.mingraviditet.extensions.userFullname
import dk.nsp.mingraviditet.ui.adapterviews.FullWidthImageViewRow
import dk.nsp.mingraviditet.ui.adapterviews.ImageViewData
import dk.nsp.mingraviditet.ui.adapterviews.LoadingState
import dk.nsp.mingraviditet.ui.adapterviews.error.ErrorStateRow
import dk.nsp.mingraviditet.ui.common.SectionHeaderAdapterView
import dk.nsp.mingraviditet.ui.main.knowledge.CategoryCard
import dk.nsp.mingraviditet.ui.main.knowledge.CategoryCardData
import dk.nsp.mingraviditet.ui.main.more.documentation.DocumentationCardData
import dk.nsp.mingraviditet.ui.main.more.documentation.DocumentationCardWithButtonRow
import dk.nsp.mingraviditet.ui.main.overview.widget.OverviewPregnancyState
import dk.nsp.mingraviditet.util.android.recycler.RecyclerAdapterView
import dk.nsp.mingraviditet.util.extensions.postSingleEvent
import dk.nsp.mingraviditet.util.extensions.toSingleEvent
import dk.nsp.mingraviditet.util.extensions.toTextModelOrDefault
import dk.nsp.mingraviditet.util.extensions.transform
import dk.nsp.mingraviditet.util.livedata.SingleEvent
import dk.nsp.models.ArticleSearchResult
import dk.nsp.models.DataCard
import dk.nsp.models.DataCardWarningState
import dk.nsp.models.DateOfBirth
import dk.nsp.shared.Failed
import dk.nsp.shared.ResultWrapper
import dk.nsp.shared.Success
import dk.nsp.shared.mapSuccess
import dk.nsp.shared.util.extensions.toDefaultMonthDisplay
import dk.nsp.shared.util.extensions.weekNumberAfterBirthDate
import dk.nsp.shared.util.sharedprefs.SharedPrefsHelper
import dk.nsp.util.searchResults
import dk.nsp.util.textmodel.TextModel
import kotlinx.coroutines.CoroutineScope
import java.time.LocalDateTime
import javax.inject.Inject

@HiltViewModel
class OverviewFragmentViewModel @Inject constructor(
    private val articlesRepository: ArticlesRepository,
    private val dataCardRepository: DataCardRepository,
    private val sharedPrefs: SharedPrefsHelper
) : ViewModel() {

    private sealed class ArticlesBirthWeekType(val week: Int) {
        class PregnantArticles(pregnancyWeek: Int) : ArticlesBirthWeekType(pregnancyWeek)
        class AfterBirthArticles(weeksAfterBirth: Int) : ArticlesBirthWeekType(weeksAfterBirth)
    }


    val shouldShowChangeDataCardButton = dataCardRepository.hasMultipleDataCards

    private val _rowsToShow = MutableLiveData<List<RecyclerAdapterView<*>>>()
    val rowsToShow: LiveData<List<RecyclerAdapterView<*>>> = _rowsToShow

    private val _onFeedbackClicked = MutableLiveData<SingleEvent<Unit>>()
    val onFeedbackClicked: LiveData<SingleEvent<Unit>> = _onFeedbackClicked

    private val _onUpdateDateOfBirthClicked = MutableLiveData<SingleEvent<DateOfBirth>>()
    val onUpdateDateOfBirthClicked: LiveData<SingleEvent<DateOfBirth>> = _onUpdateDateOfBirthClicked

    private val _onSearchedArticleClicked: MutableLiveData<SingleEvent<Pair<Article, Int>>> =
        MutableLiveData()
    val onSearchArticleClicked: LiveData<SingleEvent<Pair<Article, Int>>> =
        _onSearchedArticleClicked

    private val _navigateToKnowledge = MutableLiveData<SingleEvent<Unit>>()
    val navigateToKnowledge: LiveData<SingleEvent<Unit>> = _navigateToKnowledge

    private val _navigateToDocumentation = MutableLiveData<SingleEvent<Unit>>()
    val navigateToDocumentation: LiveData<SingleEvent<Unit>> = _navigateToDocumentation

    val usersName = dataCardRepository.currentDataCard.transform {
        it?.userFullname
            ?: dataCardRepository.userInfoFromDataCard?.name?.fullName
            ?: ""
    }

    val pregnancyProgress: LiveData<OverviewPregnancyState> =
        MediatorLiveData<OverviewPregnancyState>().apply {
            addSource(dataCardRepository.currentDataCard) {
                postValue(
                    getPregnancyState(
                        it,
                        dataCardRepository.dateOfBirthForCurrentDataCard.value
                    )
                )
            }
            addSource(dataCardRepository.dateOfBirthForCurrentDataCard) {
                postValue(
                    getPregnancyState(
                        dataCardRepository.currentDataCard.value,
                        it
                    )
                )
            }
        }

    val updateContentListener = pregnancyProgress.transform {
        viewModelScope.launchIO {
            _rowsToShow.postValue(listOf(LoadingState()))

            val pregnancyWeek =
                dataCardRepository.currentDataCard.value?.expectedDateOfDeliverySection?.week

            val weekNumberAfterBirth =
                dataCardRepository.dateOfBirthForCurrentDataCard.value?.weekNumberAfterBirthDate()
            val articlesWeek = if (weekNumberAfterBirth != null) {
                ArticlesBirthWeekType.AfterBirthArticles(weekNumberAfterBirth)
            } else {
                ArticlesBirthWeekType.PregnantArticles(pregnancyWeek ?: 1)
            }
            _rowsToShow.postValue(getRows(articlesWeek, it))
        }
    }

    val dataCardWarning: LiveData<SingleEvent<AlertDialogViewModel>> =
        MediatorLiveData<SingleEvent<AlertDialogViewModel>>().apply {
            addSource(dataCardRepository.dataCardWarning) {
                val warningState = when (it) {
                    DataCardWarningState.NoCourses -> {
                        if (sharedPrefs.dataCardNoCoursesWarningShown != true) {
                            AlertDialogViewModel(
                                title = R.string.common_no_course_available_alert_title,
                                message = R.string.common_no_course_available_alert_text,
                                cancelable = false,
                                positiveButton = AlertDialogButtonModel(R.string.common_ok) { _, _ ->
                                    sharedPrefs.dataCardNoCoursesWarningShown = true
                                }
                            )
                        } else null
                    }
                    DataCardWarningState.NoActiveCourse -> {
                        if (sharedPrefs.dataCardNoActiveCourseWarningShown != true) {
                            AlertDialogViewModel(
                                title = R.string.common_no_active_course_available_alert_title,
                                message = R.string.common_no_active_course_available_alert_text,
                                cancelable = false,
                                positiveButton = AlertDialogButtonModel(R.string.common_ok) { _, _ ->
                                    sharedPrefs.dataCardNoActiveCourseWarningShown = true
                                }
                            )
                        } else null
                    }
                    null -> null
                }
                warningState?.let {
                    postSingleEvent(it)
                }
            }
        }

    private fun onDidGiveBirthPressed() {
        _onUpdateDateOfBirthClicked.postSingleEvent(DateOfBirth())
    }

    private fun presentDateOfBirthCallback(currentDateOfBirth: LocalDateTime?) {
        _onUpdateDateOfBirthClicked.postSingleEvent(DateOfBirth(currentDateOfBirth))
    }

    private fun getPregnancyState(
        currentCard: DataCard?,
        dateOfBirth: LocalDateTime?
    ): OverviewPregnancyState =
        currentCard.getOverviewPregnancyState(
            dateOfBirth,
            ::onDidGiveBirthPressed
        ) {
            presentDateOfBirthCallback(dateOfBirth)
        }


    private suspend fun getRows(
        articlesWeekType: ArticlesBirthWeekType,
        state: OverviewPregnancyState
    ): List<RecyclerAdapterView<*>> {
        return when (state) {
            is OverviewPregnancyState.PregnancyEnded -> createEndedRows()
            else -> {
                createArticleRowsForCurrentAndPrevious(
                    articlesWeekType,
                    viewModelScope
                ) + createFeedbackCard()
            }
        }
    }

    private suspend fun createArticleRowsForCurrentAndPrevious(
        articlesWeekType: ArticlesBirthWeekType,
        scope: CoroutineScope
    ): List<RecyclerAdapterView<*>> {
        val currentWeekSearchResults = getArticleResultsForWeek(scope, articlesWeekType)
        val nextWeekSearchResults =
            getArticleResultsForWeek(scope, articlesWeekType, weekOffset = 1)

        val currentWeekItems = createArticleSection(
            currentWeekSearchResults,
            isFutureSection = false
        )

        val nextWeekItems = createArticleSection(
            nextWeekSearchResults,
            isFutureSection = true
        )

        return currentWeekItems + nextWeekItems
    }

    private fun createEndedRows() = listOf(
        FullWidthImageViewRow(ImageViewData.DrawableFromAttribute(R.attr.sofaIllustration)),
        DocumentationCardWithButtonRow(
            DocumentationCardData(
                TextModel.from(
                    R.string.overview_ended_title,
                    dataCardRepository.currentDataCard.value?.closeDateTime
                        ?.toDefaultMonthDisplay()
                        .toTextModelOrDefault()
                ),
                TextModel.from(R.string.overview_ended_description)
            )
        ),
        CategoryCard(
            CategoryCardData.WithoutIcon(
                TextModel.from(R.string.overview_ended_data_title),
                TextModel.from(R.string.overview_ended_data_description),
                TextModel.from(R.string.overview_ended_data_go_to_settings_label)
            )
            {
                _navigateToDocumentation.postSingleEvent()
            }),
        createFeedbackCard()
    )

    private fun createFeedbackCard() = OverviewFeedbackCard { _onFeedbackClicked.postSingleEvent() }

    private fun createSectionHeader(isFutureSection: Boolean): RecyclerAdapterView<*> {
        val moreButtonText =
            if (!isFutureSection) TextModel.from(R.string.overview_find_more_knowledge_button) else null
        val moreButtonAction = if (!isFutureSection) {
            { _navigateToKnowledge.postSingleEvent() }
        } else null

        return SectionHeaderAdapterView(
            TextModel.from(
                if (isFutureSection) R.string.overview_relevant_next_week_title else R.string.overview_relevant_right_now_title
            ),
            moreButtonText,
            moreButtonAction
        )
    }

    private fun createArticleSection(
        result: ResultWrapper<List<ArticleSearchResult>>,
        isFutureSection: Boolean
    ): List<RecyclerAdapterView<*>> {
        return when (result) {
            is Failed -> return listOf(
                createSectionHeader(isFutureSection),
                ErrorStateRow(
                    TextModel.from(R.string.common_alert_title_error),
                    TextModel.from("${result.error.message}")
                )
            )
            is Success -> {
                val categoryCards = result.value.map {
                    createCardForSearchResult(it)
                }

                if (categoryCards.isEmpty()) {
                    emptyList()
                } else {
                    listOf(createSectionHeader(isFutureSection)) + categoryCards
                }
            }
        }
    }

    private suspend fun getArticleResultsForWeek(
        scope: CoroutineScope,
        articlesWeekType: ArticlesBirthWeekType,
        weekOffset: Int = 0
    ): ResultWrapper<List<ArticleSearchResult>> {
        val week = articlesWeekType.week + weekOffset
        return when (articlesWeekType) {
            is ArticlesBirthWeekType.AfterBirthArticles -> articlesRepository
                .getAfterBirthArcticleCategories(scope, week).await()

            is ArticlesBirthWeekType.PregnantArticles -> articlesRepository
                .getArticleCategories(scope, week).await()
        }.mapSuccess { categories ->
            categories.flatMap { it.searchResults() }
        }
    }

    private fun createCardForSearchResult(articleSearchResult: ArticleSearchResult): CategoryCard {
        return CategoryCard(
            CategoryCardData.WithIconUrl(
                articleSearchResult.title,
                articleSearchResult.description,
                TextModel.from(articleSearchResult.subtitle),
                articleSearchResult.iconUrl
            ) {
                _onSearchedArticleClicked.postValue(
                    Pair(
                        articleSearchResult.article,
                        articleSearchResult.articleSectionId ?: -1
                    ).toSingleEvent()
                )
            })
    }
}